#! /usr/bin/env python
# ______________________________________________________________________________
"""jbmesher 
  
This script is intended to produce a Dolfin mesh for a journal bearing.

usage:
  -h    Print this usage message
  -X <val> x coordinate of the journal center
  -Y <val> y coordinate of the journal center
  -R <val> radius of journal
  -H <val> refinement of boundary. (point every sin(x+H),cos(x+H))
"""

# _____________________________________________________________________________
# Module Copyright and license
__author__ = "Andy R Terrel (aterrel@uchicago.edu)"
__date__ = "2008-09"
__copyright__ = "Copyright (C) 2008 Andy R Terrel"
__license__ = "GNU LGPL Version 3.0 or later"

# _____________________________________________________________________________
# Module Imports
import sys
import os
import getopt
from math import sqrt, pi, sin, cos

# ______________________________________________________________________________
# Module Functions
def createGEO(R,X,Y,LC,FILE):
    EPS = 1e-8
    points = []
    x = 0.0
    while x < 2*pi - LC/2 + EPS:
        points.append( (sin(x),cos(x)) )
        x += LC 
    x = 0.0
    while x < 2*pi - LC/2 + EPS:
        points.append( (R*sin(x)+X, R*cos(x)+Y) )
        x += LC 
    num_pts = len(points)

    geo_fp = open(FILE+".geo","w")
    geo_fp.write( "lc = %f;\n" % LC )

    for i,pt in enumerate(points):
        geo_fp.write("Point(%d) = {%f,%f,0.0,lc};\n" % (i+1,pt[0], pt[1]))

    # PT 1,2,3,4  5,6,7,8
    # Lines 1 (1,2); 2(2,3); 3(3,4); 4(4,1)
    #       5 (5,6); 6(6,7); 7(7,8); 8(8,5)
    #
    #
    #
    for i in xrange(1,num_pts/2):
        geo_fp.write("Line(%d) = {%d,%d};\n" % (i,i,i+1))
    geo_fp.write("Line(%d) = {%d,%d};\n" % (num_pts/2,num_pts/2,1))

    for i in xrange(num_pts/2+1,num_pts):
        geo_fp.write("Line(%d) = {%d,%d};\n" % (i,i,i+1))
    geo_fp.write("Line(%d) = {%d,%d};\n" % (num_pts,num_pts,num_pts/2+1))

    geo_fp.write("Line Loop(%d) = {%s};\n" % 
                 (num_pts+1,",".join([str(i) for i in range(1,num_pts/2+1)])))
    geo_fp.write("Line Loop(%d) = {%s};\n" % 
                 (num_pts+2,",".join([str(i) for i in range(num_pts/2+1,num_pts+1)])))

    geo_fp.write("Plane Surface(%d) = {%d,%d};\n" % (num_pts+3, num_pts+1, num_pts+2))
    geo_fp.close()

# _____________________________________________________________________________
def main(*args):
    R = 0.9
    X = 0.0
    Y = 0.0
    H = None
    try:
        opts, remainder = getopt.getopt(args, "R:X:Y:H:h")
    except getopt.GetoptError:
        print "Unrecognized options"
        print __doc__
        sys.exit(1)
    for opt_key, opt_val in opts:
        if opt_key == "-R":
            R = float(opt_val)
        elif opt_key == "-X":
            X = float(opt_val)
        elif opt_key == "-Y":
            Y = float(opt_val)
        elif opt_key == "-H":
            H = float(opt_val)
        elif opt_key == "-h":
            print __doc__
    if not H:
        LC = (1-max(R+abs(X),R+abs(Y)))/2
    else:
        LC = H

    FILE = "jb_r_%.2f_x_%.2f_y_%.2f_h_%.3f" %(R,X,Y,H)
    FILE = FILE.replace(".","_")
    createGEO(R,X,Y,H,FILE)
    os.system("gmsh %s.geo -2" % FILE)
    os.system("dolfin-convert %s.msh %s.xml" % (FILE,FILE))
    return 

# ______________________________________________________________________________
# Module Script Runner
if __name__ == "__main__":
    main(*sys.argv[1:])
# ______________________________________________________________________________
# End of mesher.py
