// Copyright (C) 2008 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.

#include <dolfin-grade2.h>

using namespace dolfin;


// ----------------------------------------------------------------------------
// Boundary Condition for outer wall
class NoSlipBC : public Function
{
public:
  NoSlipBC(Mesh& mesh) : Function(mesh) {}
  void eval (real* values, const real* x) const
  {    
      values[0] = 0; 
      values[1] = 0;
  }
};

// ----------------------------------------------------------------------------
// Boundary Condition for journal
class JournalBC : public Function
{
private:
  const real radius;
  const real x_val;
  const real y_val;
public:
  JournalBC(Mesh& mesh, const real radius, const real x_val, const real y_val):
    Function(mesh), radius(radius), x_val(x_val), y_val(y_val)
  {}
  void eval (real* values, const real* x) const
  {
    values[0] = (x[1] - y_val)/radius; 
    values[1] = -(x[0] - x_val)/radius;
  }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain
class CenterCircle : public SubDomain
{
private:
  const real radius;
  const real x_val;
  const real y_val;
public:
  CenterCircle(const real radius, const real x_val, const real y_val) : 
    SubDomain(), radius(radius), x_val(x_val), y_val(y_val)
  {
    // Do nothing
  }
  virtual bool inside (const real* x, bool on_boundary) const
  {
    double dbco = (1-sqrt(x_val*x_val+y_val*y_val)-radius);
    double dfjc = sqrt((x[0]-x_val)*(x[0]-x_val) + (x[1]-y_val)*(x[1]-y_val));
    bool ret_val = on_boundary && dfjc < radius + dbco/2;
    return on_boundary && ret_val;
  }
};

class OuterCircle : public SubDomain
{
private:
  const real radius;
  const real x_val;
  const real y_val;
public:
  OuterCircle(const real radius, const real x_val, const real y_val) : 
    SubDomain(), radius(radius), x_val(x_val), y_val(y_val)
  {
    // Do nothing
  }
  virtual bool inside (const real* x, bool on_boundary) const
  {
    double dbco = (1-sqrt(x_val*x_val+y_val*y_val)-radius);
    double dfjc = sqrt((x[0]-x_val)*(x[0]-x_val) + (x[1]-y_val)*(x[1]-y_val));
    bool ret_val = dfjc > radius + dbco/2;
    return on_boundary && ret_val;
  }
};

// ----------------------------------------------------------------------------
int main (int argc, char *argv[])
{
  char meshfile[200] = "../data/meshes/journalbearing/jb_r_0_90_x_0_00_y_-0_05_h_0_025.xml";
  real center_radius = 0.90;
  real center_x = 0.0;
  real center_y = -0.05;


  // Create the mesh, define the subsystems, and mark the boundaries.
  // by the position the bc function is in the bc func array.
  Mesh mesh(meshfile);
  CenterCircle center(center_radius,center_x,center_y);
  OuterCircle outer(center_radius,center_x,center_y);
  MeshFunction<unsigned int> vel_subdomains(mesh, mesh.topology().dim()-1);
  vel_subdomains = 2;
  center.mark(vel_subdomains,0);
  outer.mark(vel_subdomains,1);
  JournalBC   journal_bc(mesh, center_radius, center_x, center_y);
  NoSlipBC    no_slip_bc(mesh);
  Array< Function* > vel_bc_funcs(&journal_bc, &no_slip_bc);

  // RHS Function
  Function zero(mesh, 2, 0.0);
  
  // Create fluid and solver
  Fluid fluid(mesh,vel_subdomains, vel_bc_funcs);
  Grade2Solver grade2;
  grade2.set("transport disc","symmetric");
  grade2.set("max_iter", 15);
  grade2.set("alpha", 10.0);
  grade2.set("mu", 0.5);
  grade2.set("velocity disc","Scott-Vogelius");
  grade2.set("SV penalty rho", 1e6);
  grade2.set("SV penalty r", 1e6);
  grade2.set("SV max_iter",10);
  grade2.solve(fluid,zero);
  fluid.toFile("grade2");

  // Perhaps use a stokes solver instead
  StokesSolver stokes;
  stokes.solve(fluid, zero);
  fluid.toFile("stokes");

  return 0;
}

// ----------------------------------------------------------------------------
// end of file
