// Copyright (C) 2008 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.
//
// This demo implements a simple lid driven cavity with a parabolic driving
// function on the lid.

#include <dolfin-grade2.h>

using namespace dolfin;

// ----------------------------------------------------------------------------
// Boundary Subdomain marking walls
class Wall : public SubDomain
{
public:
    bool inside (const real *x, bool on_boundary) const
  {
    return on_boundary;
  }
};

// ----------------------------------------------------------------------------
// Boundary Subdomain marking lid
class Lid : public SubDomain
{
public:
    bool inside (const real *x, bool on_boundary) const
  {
    return on_boundary && (x[1] > 1.0 - DOLFIN_EPS);
  }
};
 
// ----------------------------------------------------------------------------
// Function for top flow for velocity
class TopFlow : public Function
{
public:
  
  TopFlow (Mesh& mesh) : Function(mesh) {}
  
  void eval (real* values, const real* x) const
  {
    values[0] = 4.0 * x[0] * (1.0 - x[0]);
    values[1] = 0.0;
  }
};

// -----------------------------------------------------------------------------
int main ()
{

  // Create a unit square
  UnitSquare mesh(32,32,UnitSquare::crisscross);
  Wall wall;
  Lid lid;

  // No-slip boundary condition for velocity
  Function zero(mesh, 2, 0.0);

  // Lid boundary condition for velocity
  TopFlow topflow(mesh);

  // Create a mesh function, marking boundaries by the 
  // position the bc function is in the bc func array.
  MeshFunction<unsigned int> vel_subdomains(mesh, mesh.topology().dim() - 1);
  vel_subdomains = 2;
  wall.mark(vel_subdomains,0);
  lid.mark(vel_subdomains,1);
  Array< Function* > vel_bc_funcs(&zero, &topflow);

  // Create fluid and solver
  Fluid fluid(mesh,vel_subdomains, vel_bc_funcs);
  Grade2Solver grade2;
  grade2.set("max_iter", 15);
  grade2.set("alpha", 0.1);
  grade2.set("mu", 10.0);
  grade2.solve(fluid,zero);
  fluid.toFile("grade2");

  // Perhaps use a stokes solver instead
  StokesSolver stokes;
  stokes.set("velocity disc","Scott-Vogelius");
  stokes.set("SV max_iter",10);
  stokes.solve(fluid, zero);
  fluid.toFile("stokes");

  return 0;
}

// ----------------------------------------------------------------------------
// end of file
