// Copyright (C) 2008 Andy R Terrel.
// Licensed under the GNU LGPL Version 3.0 or later.
//
// -----------------------------------------------------------------------------
//  grade2_test.cpp
//
//  This program provides a series of tests using the g2 model. For the
//  test problems see test_problems.h
//
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// include statements
// -----------------------------------------------------------------------------
#include <dolfin.h>

//Included Forms for Grade 2
#include <forms/grade2/TaylorHood/P2THG2Vel.h>
#include <forms/grade2/TaylorHood/P2THG2GalZ.h>
#include <forms/grade2/TaylorHood/P2THG2SymZ.h>
#include <forms/grade2/TaylorHood/P2THG2StrZ.h>
#include <forms/grade2/TaylorHood/P2THG2SymStrZ.h>
#include <forms/grade2/TaylorHood/P2THG2RegZ.h>

// Included forms for various norms
#include <forms/norms/P2VecL2.h>
#include <forms/norms/P2VecL2Diff.h>
#include <forms/norms/P2VecH1Diff.h>
#include <forms/norms/P2L2.h>
#include <forms/norms/P2L2Diff.h>

// Include the test problems
#include "../test_problems.h"
#include "../options.h"

// -----------------------------------------------------------------------------
// global declarations
// -----------------------------------------------------------------------------
using namespace dolfin;

using dolfin::uint;

// -----------------------------------------------------------------------------
void runTHSolveLoop(Mesh mesh, Function &u, Function &p, Function &z,
		    LinearPDE &pde_vform, LinearPDE &pde_zform, Function &u_ex, 
		    Function &p_ex, Function &z_ex,
		    Options *options)
{
  P2VecH1DiffFunctional m_H1u(u,u_ex);
  P2VecL2DiffFunctional m_L2u(u,u_ex);
  P2L2DiffFunctional m_L2z(z,z_ex);
  for(int z_iters = 0; z_iters<options->max_iters; z_iters++) {
    pde_vform.solve(u,p);
    pde_zform.solve(z);
    fprintf(options->out_fp,"Iteration %d\n  H1 Error u: %.2e\n  L2 Error u: %.2e\n  L2 Error z: %.2e\n",
	    z_iters,sqrt(assemble(m_H1u,mesh)), sqrt(assemble(m_L2u,mesh)),
	    sqrt(assemble(m_L2z,mesh)));
  }
}

// -----------------------------------------------------------------------------
void runSim(Options* options)
{
  Mesh* mesh;
  if (options->problem == 6) {
    mesh = new Rectangle(0,DOLFIN_PI,0,DOLFIN_PI,options->nx,options->ny,
			 Rectangle::left);
  } else {
    mesh = new UnitSquare(options->nx,options->ny,UnitSquare::crisscross);
  }
  MeshFunction<unsigned int> vel_sub_domains(*mesh, mesh->topology().dim()-1);
  vel_sub_domains = 1;

  TestBoundary outer_boundary;
  TestPenPoint      penpoint;
  outer_boundary.mark(vel_sub_domains,0);
  
  // Create functions for boundary conditions
  TestVelocity    velocity_bc(*mesh,options->problem);
  TestPressureBC  pressure_bc(*mesh);
  SubSystem   velocity(0);
  SubSystem   pressure(1);

  // Boundary condition for velocity
  DirichletBC bc0(velocity_bc, vel_sub_domains, 0, velocity);
  DirichletBC bc1(pressure_bc, *mesh, penpoint, pressure, pointwise);

  // Boundary condition for z
  TestZ z_bc(*mesh, options->problem, options->alpha, options->mu);
  MeshFunction<unsigned int> z_sub_domain(*mesh,mesh->topology().dim()-1);
  z_sub_domain = 1;
  outer_boundary.mark(z_sub_domain,0);
  DirichletBC bcz(z_bc, z_sub_domain,0);

  TestRHS f(*mesh, options->problem, options->alpha, options->mu);
  Array<DirichletBC *> bcs(&bc0,&bc1);
  
  Function z(*mesh,0.0);
  Function u,p;
  Function F_mu(*mesh,options->mu);
  Function F_alpha(*mesh,options->alpha);

  P2THG2VelBilinearForm a_vform(z,F_mu);
  P2THG2VelLinearForm L_vform(f);

  if (options->zdisc == "Galerkin") {
    P2THG2GalZBilinearForm a_zform(u,F_mu,F_alpha);
    P2THG2GalZLinearForm L_zform(u, F_mu,F_alpha, f);
    LinearPDE pde_vform(a_vform,L_vform, *mesh, bcs);
    LinearPDE pde_zform(a_zform,L_zform, *mesh, bcz);
    runTHSolveLoop(*mesh, u, p, z, pde_vform, pde_zform,
		   velocity_bc, z_bc, pressure_bc, options);
  }
//   case SYMMETRIC: {
//     P2THG2SymZBilinearForm a_zform(u,F_mu,F_alpha);
//     P2THG2SymZLinearForm L_zform(u,F_mu,F_alpha, f);
//     LinearPDE pde_vform(a_vform,L_vform, mesh, bcs);
//     LinearPDE pde_zform(a_zform,L_zform, mesh);
//     runTHSolveLoop(mesh, u, p, z, pde_vform, pde_zform, options);
//     break;
//   }
//   case STREAMLINE: {
//     Function F_stream(mesh,options->streamline_param);
//     P2THG2StrZBilinearForm a_zform(u,F_mu,F_alpha,F_stream);
//     P2THG2StrZLinearForm L_zform(u,F_mu,F_alpha,f,F_stream);
//     LinearPDE pde_vform(a_vform,L_vform, mesh, bcs);
//     LinearPDE pde_zform(a_zform,L_zform, mesh);
//     runTHSolveLoop(mesh, u, p, z, pde_vform, pde_zform, options);
//     break;
//   }
//   case SYMMETRIC_STREAMLINE: {
//     Function F_stream(mesh,options->streamline_param);
//     P2THG2SymStrZBilinearForm a_zform(u,F_mu,F_alpha,F_stream);
//     P2THG2SymStrZLinearForm L_zform(u,F_mu,F_alpha,f,F_stream);
//     LinearPDE pde_vform(a_vform,L_vform, mesh, bcs);
//     LinearPDE pde_zform(a_zform,L_zform, mesh);
//     runTHSolveLoop(mesh, u, p, z, pde_vform, pde_zform, options);
//     break;
//   }
//   case REGULARIZED: {
//     Function F_regular(mesh,options->regularize_param);
//     FacetNormal n(mesh);
//     P2THG2RegZBilinearForm a_zform(u,F_mu,F_alpha,F_regular,n);
//     P2THG2RegZLinearForm L_zform(u,F_mu,F_alpha,f);
//     LinearPDE pde_vform(a_vform,L_vform, mesh, bcs);
//     LinearPDE pde_zform(a_zform,L_zform, mesh);
//     runTHSolveLoop(mesh, u, p, z, pde_vform, pde_zform, options);
//     break;
//   }  default: 
//     message("Z Discretiztion not implemented\n");
//     std::exit(1);
//   }
}


// -----------------------------------------------------------------------------
void runComparison(Options* options)
{
//   UnitSquare mesh(options->nx,options->ny);
//   MeshFunction<unsigned int> vel_sub_domains(mesh, mesh.topology().dim()-1);
//   vel_sub_domains = 0;    
//   MeshFunction<unsigned int> pres_sub_domains(mesh, mesh.topology().dim()-1);
//   pres_sub_domains = 0;    

//   TestBoundary outer_boundary;
//   TestPenPoint      penpoint;
//   outer_boundary.mark(vel_sub_domains,1);
//   penpoint.mark(pres_sub_domains,1);
  
//   // Create functions for boundary conditions
//   TestVelocity    velocity_bc(mesh,options->problem);
//   TestPressureBC  pressure_bc(mesh);
//   SubSystem   velocity(0);
//   SubSystem   pressure(1);

//   // Boundary condition for velocity
//   DirichletBC bc0(velocity_bc, vel_sub_domains, 1, velocity);
//   DirichletBC bc1(pressure_bc, pres_sub_domains, 1, pressure);

//   TestRHS f(mesh, options->problem, options->alpha, options->mu);
//   Array<DirichletBC *> bcs(&bc0,&bc1);
  
//   Function z_gal(mesh,0.0), u_gal, p_gal;
//   Function z_sym(mesh,0.0), u_sym, p_sym;
//   Function z_str(mesh,0.0), u_str, p_str;
//   Function z_symstr(mesh,0.0), u_symstr, p_symstr;
//   Function z_reg(mesh,0.0), u_reg, p_reg;
//   Function F_mu(mesh,options->mu);
//   Function F_alpha(mesh,options->alpha);
//   Function F_stream(mesh,options->streamline_param);
//   Function F_regular(mesh,options->regularize_param);

//   P2THG2VelLinearForm L_vform(f);

//   // case GALERKIN:
//   fprintf(options->out_fp, "Galerkin Z Discretization\n");
//   P2THG2VelBilinearForm a_gal_vform(z_gal,F_mu);
//   LinearPDE pde_gal_vform(a_gal_vform,L_vform, mesh, bcs);
//   P2THG2GalZBilinearForm a_z_gal_form(u_gal,F_mu,F_alpha);
//   P2THG2GalZLinearForm L_z_gal_form(u_gal,F_mu,F_alpha, f);
//   LinearPDE pde_z_gal_form(a_z_gal_form,L_z_gal_form, mesh);
//   runTHSolveLoop(mesh, u_gal, p_gal, z_gal, pde_gal_vform, pde_z_gal_form, options);

//   //  case SYMMETRIC:
//   fprintf(options->out_fp, "Symmetric Z Discretization\n");
//   P2THG2VelBilinearForm a_sym_vform(z_sym,F_mu);
//   LinearPDE pde_sym_vform(a_sym_vform,L_vform, mesh, bcs);
//   P2THG2SymZBilinearForm a_z_sym_form(u_sym,F_mu,F_alpha);
//   P2THG2SymZLinearForm L_z_sym_form(u_sym,F_mu,F_alpha, f);
//   LinearPDE pde_z_sym_form(a_z_sym_form,L_z_sym_form, mesh);
//   runTHSolveLoop(mesh, u_sym, p_sym, z_sym, pde_sym_vform, pde_z_sym_form, options);

//   // case STREAMLINE:
//   fprintf(options->out_fp, "Streamline Z Discretization\n");
//   P2THG2VelBilinearForm a_str_vform(z_str,F_mu);
//   LinearPDE pde_str_vform(a_str_vform,L_vform, mesh, bcs);
//   P2THG2StrZBilinearForm a_z_str_form(u_str,F_mu,F_alpha, F_stream);
//   P2THG2StrZLinearForm L_z_str_form(u_str,F_mu,F_alpha, f, F_stream);
//   LinearPDE pde_z_str_form(a_z_str_form,L_z_str_form, mesh);
//   runTHSolveLoop(mesh, u_str, p_str, z_str, pde_str_vform, pde_z_str_form, options);
  
//   // case SYMMETRIC_STREAMLINE:
//   fprintf(options->out_fp, "Symmetric Streamline Z Discretization\n");
//   P2THG2VelBilinearForm a_symstr_vform(z_symstr,F_mu);
//   LinearPDE pde_symstr_vform(a_symstr_vform,L_vform, mesh, bcs);
//   P2THG2SymStrZBilinearForm a_z_symstr_form(u_symstr,F_mu,F_alpha, F_stream);
//   P2THG2SymStrZLinearForm L_z_symstr_form(u_symstr,F_mu,F_alpha, f, F_stream);
//   LinearPDE pde_z_symstr_form(a_z_symstr_form,L_z_symstr_form, mesh);
//   runTHSolveLoop(mesh, u_symstr, p_symstr, z_symstr, pde_symstr_vform, pde_z_symstr_form, options);

//   // case REGULARIZED:
//   fprintf(options->out_fp, "Regularized Z Discretization\n");
//   FacetNormal n(mesh);
//   P2THG2VelBilinearForm a_reg_vform(z_reg,F_mu);
//   LinearPDE pde_reg_vform(a_reg_vform,L_vform, mesh, bcs);
//   P2THG2RegZBilinearForm a_z_reg_form(u_reg,F_mu,F_alpha, F_regular,n);
//   P2THG2RegZLinearForm L_z_reg_form(u_reg,F_mu,F_alpha, f);
//   LinearPDE pde_z_reg_form(a_z_reg_form,L_z_reg_form, mesh);
//   runTHSolveLoop(mesh, u_reg, p_reg, z_reg, pde_reg_vform, pde_z_reg_form, options);
  
//   TestVelocity u(mesh, options->problem);
//   real gal_u_error,sym_u_error,str_u_error,symstr_u_error, reg_u_error;
//   P2VecH1DiffFunctional m_gal_error(u_gal,u);
//   gal_u_error = sqrt(assemble(m_gal_error,mesh));
//   P2VecH1DiffFunctional m_sym_error(u_sym,u);
//   sym_u_error = sqrt(assemble(m_sym_error,mesh));
//   P2VecH1DiffFunctional m_str_error(u_str,u);
//   str_u_error = sqrt(assemble(m_str_error,mesh));
//   P2VecH1DiffFunctional m_symstr_error(u_symstr,u);
//   symstr_u_error = sqrt(assemble(m_symstr_error,mesh));
//   P2VecH1DiffFunctional m_reg_error(u_reg,u);
//   reg_u_error = sqrt(assemble(m_reg_error,mesh));

//   P2VecH1DiffFunctional m_gal_sym(u_gal,u_sym);
//   real gal_sym = sqrt(assemble(m_gal_sym,mesh));
//   P2VecH1DiffFunctional m_gal_str(u_gal,u_str);
//   real gal_str = sqrt(assemble(m_gal_str,mesh));
//   P2VecH1DiffFunctional m_gal_symstr(u_gal,u_symstr);
//   real gal_symstr = sqrt(assemble(m_gal_symstr,mesh));
//   P2VecH1DiffFunctional m_gal_reg(u_gal,u_reg);
//   real gal_reg = sqrt(assemble(m_gal_reg,mesh));

//   P2VecH1DiffFunctional m_sym_str(u_sym,u_str);
//   real sym_str = sqrt(assemble(m_sym_str,mesh));
//   P2VecH1DiffFunctional m_sym_symstr(u_sym,u_symstr);
//   real sym_symstr = sqrt(assemble(m_sym_symstr,mesh));
//   P2VecH1DiffFunctional m_sym_reg(u_sym,u_reg);
//   real sym_reg = sqrt(assemble(m_sym_reg,mesh));

//   P2VecH1DiffFunctional m_str_symstr(u_str,u_symstr);
//   real str_symstr = sqrt(assemble(m_str_symstr,mesh));
//   P2VecH1DiffFunctional m_str_reg(u_str,u_reg);
//   real str_reg = sqrt(assemble(m_str_reg,mesh));

//   P2VecH1DiffFunctional m_symstr_reg(u_symstr,u_reg);
//   real symstr_reg = sqrt(assemble(m_symstr_reg,mesh));
  
//   fprintf(options->out_fp, "U H1 Error (Problem %d, Max Iters %d, alpha %.2e, mu %.2e) \\\\\n",
// 	  options->problem, options->max_iters, options->alpha, options->mu);
//   fprintf(options->out_fp,"Galerkin & %.2e \\\\\nSymmetric & %.2e \\\\\n",
// 	  gal_u_error,sym_u_error);
//   fprintf(options->out_fp, "Streamline & %.2e \\\\\nSymmetricStreamline & %.2e \\\\\nRegularized & %.2e \\\\\n",
// 	  str_u_error,symstr_u_error,reg_u_error);
//   fprintf(options->out_fp, "U H1 Diffs (Problem %d, Max Iters %d, alpha %.2e, mu %.2e) \\\\\n",
// 	  options->problem, options->max_iters, options->alpha, options->mu);
//   fprintf(options->out_fp,"                     & Symmetric & Streamline & Symmetric Streamline & Regularized\\\\\n");
//   fprintf(options->out_fp,"Galerkin             & %.2e      & %.2e       & %.2e                 & %.2e \\\\\n",
// 	  gal_sym, gal_str, gal_symstr, gal_reg);
//   fprintf(options->out_fp,"Symmetric            & --        & %.2e       & %.2e                 & %.2e \\\\\n",
// 	  sym_str, sym_symstr, sym_reg);
//   fprintf(options->out_fp,"Streamline           & --        & --         & %.2e                 & %.2e \\\\\n",
// 	  str_symstr, str_reg);
//   fprintf(options->out_fp,"Symmetric Streamline & --        & --         & --                   & %.2e \\\\\n",
// 	  symstr_reg);
}
// -----------------------------------------------------------------------------
// Main function
// -----------------------------------------------------------------------------
int main (int argc, char *argv[])
{
  Options options;
  processOptions(argc,argv,&options);
  if (options.run_comparison) {
    runComparison(&options);
  } else {
    runSim(&options);
  }
}

// -----------------------------------------------------------------------------
// end of file grade2_test.cpp
