// Copyright (C) 2008 Andy R Terrel.
// Licensed under the GNU LGPL Version 3.0 or later.
//
// ----------------------------------------------------------------------------
//  grade2_test.cpp
//
//  This program provides a series of tests using the g2 model. For the
//  test problems see test_problems.h
//
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
// include statements
// ----------------------------------------------------------------------------
#include <dolfin-grade2.h>

// Included forms for various norms
//#include <forms/norms/P2VecL2.h>
#include <forms/norms/P2VecH1Diff.h>
//#include <forms/norms/P2L2.h>
//#include <forms/norms/P2L2Diff.h>
#include "P4VecH1Diff.h"

// Include the test problems
#include "../test_problems.h"

// Include options parser
#include "../options.h"

// ----------------------------------------------------------------------------
// global declarations
// ----------------------------------------------------------------------------
using namespace dolfin;

void computeError(Fluid& fluid, Options* options)
{
  TestVelocity u(fluid.mesh(), options->problem);      
  real h1_error;
  Form* m_error = 0;
  if (options->vdisc == "Scott-Vogelius") {
    m_error = new P4VecH1DiffFunctional(fluid.velocity(),u);
  } else if (options->vdisc == "Taylor-Hood") {
    m_error = new P2VecH1DiffFunctional(fluid.velocity(),u);
  } else {
    cout << "Unable to take error of discretization";
  }
  h1_error = sqrt(fabs(assemble(*m_error,fluid.mesh())));
  std::cout << "Velocity H1 Error: ";
  std::cout << std::setiosflags(std::ios_base::scientific)
	    << std::setprecision(2)
	    << h1_error << std::endl;

  delete m_error;
}
// ----------------------------------------------------------------------------
void runSim(Options* options)
{
  Mesh* mesh;
  if (options->problem == 6) {
    mesh = new Rectangle(0,DOLFIN_PI,0,DOLFIN_PI,options->nx,options->ny,
			 Rectangle::left);
  } else {
    mesh = new UnitSquare(options->nx,options->ny,UnitSquare::crisscross);
  }
  MeshFunction<unsigned int> vel_subdomains(*mesh, mesh->topology().dim()-1);
  vel_subdomains = 1;
  TestBoundary outer_boundary;
  outer_boundary.mark(vel_subdomains,0);
  TestVelocity velocity_ex(*mesh,options->problem);
  TestPressure pressure_ex(*mesh,options->problem);
  TestZ z_ex(*mesh, options->problem, options->alpha, options->mu);
  Array< Function* > vel_bc_funcs;
  vel_bc_funcs.push_back(&velocity_ex);
  
  TestRHS f(*mesh, options->problem, options->alpha, options->mu);
  
  Fluid fluid(*mesh, vel_subdomains, vel_bc_funcs);
  Grade2Solver grade2;

  grade2.set("alpha", options->alpha);
  grade2.set("mu", options->mu);
  grade2.set("upwind param",options->streamline_param);
  grade2.set("regularize param",options->regularize_param);
  grade2.set("max_iter",options->max_iters);
  

  if (not options->run_comparison) {
    grade2.set("velocity disc", options->vdisc);
    grade2.set("transport disc", options->zdisc);
    grade2.solve(fluid, f, velocity_ex, pressure_ex, z_ex);
    computeError(fluid, options);
  } else {
    for (int i=0; i<5;i++) {
      for (int j=0; j<2;i++) {	
	grade2.set("transport disc", ZDiscStr[i]);
	grade2.set("velocity disc", VDiscStr[j]);
	grade2.solve(fluid, f, velocity_ex, pressure_ex, z_ex);
	computeError(fluid, options);
      }
    }
  }

  delete mesh;
}
// ----------------------------------------------------------------------------
// Main function
// ----------------------------------------------------------------------------
int main (int argc, char *argv[])
{
  Options options;
  processOptions(argc,argv,&options);
  runSim(&options);  
}
// ----------------------------------------------------------------------------
// end of file
