// Copyright (C) 2008 Andy R Terrel.
// Licensed under the GNU LGPL Version 3.0 or later.
//
// -----------------------------------------------------------------------------
//  grade2_test.cpp
//
//  This program provides a series of tests using the g2 model. For the
//  test problems see test_problems.h
//
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------
// include statements
// -----------------------------------------------------------------------------
#include <dolfin-grade2.h>

// Included forms for various norms
#include <forms/norms/P2VecL2.h>
#include <forms/norms/P2VecH1Diff.h>
#include <forms/norms/P2L2.h>
#include <forms/norms/P2L2Diff.h>
#include "P2P15L2Diff.h"

// Include the test problems
#include "../test_problems.h"

// Include option parser
#include "../options.h"

// -----------------------------------------------------------------------------
// global declarations
// -----------------------------------------------------------------------------
using namespace dolfin;

using dolfin::uint;

// -----------------------------------------------------------------------------
void runSim(Options* options)
{

  real alpha[4] = {0.0,0.1,1.0,-0.1};
  real perturb[3] = {0.0, 1e-4, 1e-2};
  int mesh_size[6] = {4,8,16,32,64,128};
  std::string method[2] = {"Galerkin","regularized"};
  real norms[4][3][6][2];
  real l2error[4][3][6][2];

  for (int i_alpha = 0; i_alpha < 4; i_alpha++) {
    for (int i_mesh_size = 0; i_mesh_size < 6; i_mesh_size++) {
      int n = mesh_size[i_mesh_size];
      Mesh* mesh;
      if (options->problem == 6) {
	mesh = new Rectangle(0,DOLFIN_PI,0,DOLFIN_PI,n,n,
			     Rectangle::left);
      } else {
	mesh = new UnitSquare(n,n,UnitSquare::crisscross);
      }
      for (int i_perturb = 0; i_perturb < 3; i_perturb++) {
	real delta = perturb[i_perturb];

	TestVelocity u(*mesh, options->problem, delta, -delta);
	Function Falpha(*mesh,alpha[i_alpha]);
	Function mu(*mesh, options->mu);
	Function reg(*mesh, options->regularize_param / n);
	FacetNormal normal(*mesh);
	TestRHS rhs(*mesh,options->problem,alpha[i_alpha],options->mu);
	TestZ z(*mesh, options->problem, alpha[i_alpha], options->mu);
    
	P2THG2GalZBilinearForm a_zform(u, mu, Falpha);
	P2THG2GalZLinearForm L_zform(u, mu, Falpha, rhs);
	LinearPDE pde_zform(a_zform,L_zform,*mesh);
	Function z_h;
	pde_zform.solve(z_h);
      
	P2P15L2DiffFunctional m_L2z(z_h,z);
	l2error[i_alpha][i_perturb][i_mesh_size][0] = sqrt(assemble(m_L2z, *mesh));
	
	P2L2Functional m_L2znorm(z_h);
	norms[i_alpha][i_perturb][i_mesh_size][0] = sqrt(assemble(m_L2znorm, *mesh));

	P2THG2RegZBilinearForm a_reg_zform(u, mu, Falpha, reg, normal);
	P2THG2RegZLinearForm L_reg_zform(u, mu, Falpha, rhs);
	LinearPDE pde_reg_zform(a_reg_zform,L_reg_zform,*mesh);
	pde_reg_zform.solve(z_h);
  
	l2error[i_alpha][i_perturb][i_mesh_size][1] = sqrt(assemble(m_L2z, *mesh));
	norms[i_alpha][i_perturb][i_mesh_size][1] = sqrt(assemble(m_L2znorm, *mesh));
      }
      delete mesh;
    }
  }
  
  fprintf(options->out_fp,"Alpha, perturb, method, mesh, ||z_h||, ||z_h - z||\\\\\\hline\n");
  for (int i_alpha = 0; i_alpha < 4; i_alpha++) {
    for (int i_perturb = 0; i_perturb < 3; i_perturb++) {      
      for (int i_method = 0; i_method < 2; i_method++) {
	for (int i_mesh_size = 0; i_mesh_size < 6; i_mesh_size++) {
	  fprintf(options->out_fp, "%.2e & %.2e & %s & %dX%d & %.2e & %.2e \\\\\\hline\n",
		  alpha[i_alpha], perturb[i_perturb], method[i_method].c_str(), mesh_size[i_mesh_size],
		  mesh_size[i_mesh_size],
		  norms[i_alpha][i_perturb][i_mesh_size][i_method],
		  l2error[i_alpha][i_perturb][i_mesh_size][i_method]);
	}
      }
    }
  }
    

//     File outfile("proj_test.pvd");
//     outfile << z_h;
//     outfile << z;


//   TestBoundary outer_boundary;
//   TestPenPoint      pinpoint;
//   outer_boundary.mark(vel_sub_domains,1);
  
//   // Create functions for boundary conditions
//   TestVelocity    velocity_bc(mesh,options->problem);
//   TestPressureBC  pressure_bc(mesh);
//   SubSystem   velocity(0);
//   SubSystem   pressure(1);

//   // Boundary condition for velocity
//   DirichletBC bc0(velocity_bc, vel_sub_domains, 1, velocity);
//   DirichletBC bc1(pressure_bc, mesh, pinpoint, pressure, pointwise);

//   TestRHS f(mesh, options->problem, options->alpha, options->mu);
//   Array<DirichletBC *> bcs(&bc0,&bc1);
  
//   TestZ z(mesh,options->problem,options->alpha,options->mu);
//   Function u,p;
//   Function F_mu(mesh,options->mu);
//   Function F_alpha(mesh,options->alpha);

//   P2THG2VelBilinearForm a_vform(z,F_mu);
//   P2THG2VelLinearForm L_vform(f);
//   LinearPDE pde_vform(a_vform,L_vform, mesh, bcs);
//   pde_vform.solve(u,p);

//   TestVelocity u_ex(mesh,options->problem);
//   P2VecH1DiffFunctional m_error(u,u_ex);
//   real h1_error = sqrt(fabs(assemble(m_error,mesh)));
//   std::cout << "Velocity H1 Error: ";
//   std::cout << std::setiosflags(std::ios_base::scientific)
// 	    << std::setprecision(2)
// 	    << h1_error << std::endl;

}

// -----------------------------------------------------------------------------
// Main function
// -----------------------------------------------------------------------------
int main (int argc, char *argv[])
{
  Options options;
  processOptions(argc,argv,&options);
  runSim(&options);
}

// -----------------------------------------------------------------------------
// end of file grade2_test.cpp
