using namespace dolfin;


using dolfin::uint;

enum ZDiscType { GALERKIN,
		 SYMMETRIC,
		 STREAMLINE,
		 SYMMETRIC_STREAMLINE,
		 REGULARIZED };
const char* ZDiscStr[] = {"Galerkin",
			  "symmetric",
			  "streamline",
			  "symmetric streamline",
			  "regularized" };

const char* VDiscStr[] = {"Taylor-Hood",
			  "Scott-Vogelius" };

// Top level data definitions
typedef struct {
  int      debug;                 // The debugging level
  int      problem;               // Case 1 -- 4
  char     output[2048];          // The filename to output some results
  char     base_vtk[2048];        // The base filename to write out vtk files
  FILE*    out_fp;                // Output file pointer
  int      nx,ny;                 // Number of segment in x and y
  real     mu,alpha;              // Rheology parameters
  int      max_iters;             // Number of iterations in the uz loop
  std::string   zdisc;                 // The type of z discretization
  std::string   vdisc;                 // The type of z discretization
  real     streamline_param;      // The streamline parameter
  real     regularize_param;      // The regularizing parameter
  bool     run_comparison;        // Run a comparison of the discretizations
} Options;
// ----------------------------------------------------------------------------
// Function Definitions
// ----------------------------------------------------------------------------
void processOptions (int argc, char* argv[], Options* options)
{
  //Setup Defaults
  options->debug         = 0;
  options->problem       = 1;
  options->nx            = 16;
  options->ny            = 16;
  options->mu            = 1.0;
  options->alpha         = 1.0;
  options->zdisc         = "Galerkin";
  options->vdisc         = "Taylor-Hood";
  options->max_iters     = 20;
  options->streamline_param = 0.5;
  options->regularize_param = 1.0e-3;
  options->run_comparison = false;
  strcpy(options->output,"grade2_test.out");
  strcpy(options->base_vtk,"");

  for(int i=1; i<argc; i++) {
    if (!strcmp("--problem",argv[i])) {
      options->problem = atoi(argv[++i]);
    } else if (!strcmp("--nx",argv[i])) {
      options->nx = atoi(argv[++i]);
    } else if (!strcmp("--ny",argv[i])) {
      options->ny = atoi(argv[++i]);
    } else if (!strcmp("--max_iters",argv[i])) {
      options->max_iters = atoi(argv[++i]);
    } else if (!strcmp("--mu",argv[i])) {
      options->mu = atof(argv[++i]);
    } else if (!strcmp("--alpha",argv[i])) {
      options->alpha = atof(argv[++i]);
    } else if (!strcmp("--regularize",argv[i])) {
      options->regularize_param = atof(argv[++i]);
    } else if (!strcmp("--streamline",argv[i])) {
      options->streamline_param = atof(argv[++i]);
    } else if (!strcmp("--debug",argv[i])) {
      options->debug = atoi(argv[++i]);
    } else if (!strcmp("--output",argv[i])) {
      strcpy(options->output,argv[++i]);
    } else if (!strcmp("--base_vtk",argv[i])) {
      strcpy(options->base_vtk,argv[++i]);
    } else if (!strcmp("--run_comparison",argv[i])) {
      options->run_comparison = true;
    } else if (!strcmp("--vdisc",argv[i])) {
      options->vdisc = argv[++i];
    } else if (!strcmp("--zdisc",argv[i])) {
      options->zdisc = argv[++i];
    } else {
      message("Unprocessed option: %s", argv[i]);
    }
  }
  options->out_fp = fopen(options->output,"a");
  if (options->problem < 1 || options->problem > 6) {
    printf("Please choose a valid problem (1 -- 6)\n");
    std::exit(1);
  }
  fprintf(options->out_fp,"\nStarting simulation run with following options:\n");
  fprintf(options->out_fp,"\tMesh Segments: (%d,%d)\n", options->nx,options->ny);
  fprintf(options->out_fp,"\tProblem: %d\n", options->problem);
  if (!options->run_comparison) {
    fprintf(options->out_fp,"\tv discretization: %s\n", options->vdisc.c_str());
    fprintf(options->out_fp,"\tz discretization: %s\n", options->zdisc.c_str());
  } else {
    fprintf(options->out_fp,"\tz discretization: all\n");
  }
  if (options->zdisc == "streamline" || options->zdisc == "symstr")
    fprintf(options->out_fp,"\tStreamline param: %f\n", 
	    options->streamline_param);
  else if (options->zdisc == "regularized")
    fprintf(options->out_fp,"\tRegularize param: %f\n", 
	    options->regularize_param);
  fprintf(options->out_fp,"\tmu: %f\n", options->mu);
  fprintf(options->out_fp,"\talpha: %f\n",options->alpha);
  fprintf(options->out_fp,"\tdebug: %d\n",options->debug);
  fprintf(options->out_fp,"\toutput: %s\n",options->output);
  fprintf(options->out_fp,"\tbase_vtk: %s\n",options->base_vtk);
}
// ----------------------------------------------------------------------------
