// Copyright (C) 2008 Andy R Terrel.
// Licensed under the GNU LGPL Version 3.0 or later.
//
// First added:  2008-09-09
// Last changed: 2008-09-09
// -----------------------------------------------------------------------------
//  The following test problems are implemented.
//  Case 1) [0,1]X[0,1] Stokes equations with solution: 
//            u[0] =  sin(3*pi*x) * cos(3*pi*y)
//            u[1] = -cos(3*pi*x) * sin(3*pi*y)
//            p    =  sin(3*pi*x) * sin(3*pi*y)
//
//  Case 2) [0,1]X[0,1] Stokes equations with solution: 
//            u[0] =  x^2(1-x)^2(2y-6y^2+4y^3)
//            u[1] = -y^2(1-y)^2(2x-6x^2+4x^3)
//            p    =  x(1-x)
//
//  Case 3) [0,1]X[0,1] Stokes equations with solution: 
//            u[0] =  sin(12*pi*x) * cos(12*pi*y)
//            u[1] = -cos(12*pi*x) * sin(12*pi*y)
//            p    =  sin(12*pi*x) * sin(12*pi*y)
//
//  Case 4) [0,1]X[0,1] Grade two equation with:
//            u = curl psi, z = curl(u - alpha*delta(u))
//            psi(x,y) = (y * (1 - y) * sin pi*x)^2
//            p(x,y) = cos(pi*x) * cos(2*pi*y) + c
//
// -----------------------------------------------------------------------------

#ifndef __TEST_PROBLEMS_H
#define __TEST_PROBLEMS_H

#include <dolfin.h>

namespace dolfin
{
  class D2R1Function: public Function
  {
  public:
  D2R1Function(Mesh& mesh) : Function(mesh) {}
    virtual uint rank() const {return 1;}
    virtual uint dim(unsigned int i) const {return 2;}
  };

  // -----------------------------------------------------------------------------
  class D1R1Function: public Function
  {
  public:
  D1R1Function(Mesh& mesh) : Function(mesh) {}
    virtual uint rank() const {return 1;}
    virtual uint dim(unsigned int i) const {return 1;}
  };

  // -----------------------------------------------------------------------------
  class D1R0Function: public Function
  {
  public:
  D1R0Function(Mesh& mesh) : Function(mesh) {}
    virtual uint rank() const {return 0;}
    virtual uint dim(unsigned int i) const {return 1;}
  };

  // -----------------------------------------------------------------------------
  class TestPressure: public Function
  {
  private:
    int _problem;
  public:
  TestPressure(Mesh& mesh,int problem) : Function(mesh),_problem(problem)
    {}
    uint rank() const {return 0;}
    real eval(const real* x) const
    {
      real ret_val;
      switch (_problem) {
      case 1:
	ret_val=sin(3*DOLFIN_PI*x[0]) * sin(3*DOLFIN_PI*x[1]);
	break;
      case 2:
	ret_val= x[0]*(1.0-x[0]);
	break;
      case 3:
	ret_val=sin(12*DOLFIN_PI*x[0]) * sin(12*DOLFIN_PI*x[1]);
	break;
      case 4:
	ret_val=cos(x[0]) * cos(2*x[1]);
	break;
      case 5:
	ret_val=x[0] * x[1];
	break;
      case 6:
	ret_val=cos(x[0]) * cos(2*x[1]);
	break;
      default:
	message("Unknown problem in Pressure, given %d",_problem);
	std::exit(1);
      }
      return ret_val;
    }
  };

  // -----------------------------------------------------------------------------
  class TestVelocity : public D2R1Function
  {
  private:
    int _problem;
    real _perturb_x;
    real _perturb_y;
  public:
    TestVelocity(Mesh& mesh,int problem,real perturb_x = 0.0, real perturb_y = 0.0) 
    : D2R1Function(mesh),_problem(problem), 
	_perturb_x(perturb_x), _perturb_y(perturb_y)
    {}
    void eval(real* values, const real* input) const
    {
      real pi = DOLFIN_PI;
      real x = input[0];
      //real x2 = x*x;
      //real x3 = x2*x;
      real y = input[1];
      real y2 = y*y;
      //real y3 = y2*y;
      switch (_problem) {
      case 1:
	values[0] = sin(3*pi*x) * cos(3*pi*y);
	values[1] = -cos(3*pi*x) * sin(3*pi*y);
	break;
      case 2:
	values[0] = x*x*(1.0-x)*(1.0-x)
	  *(2.0*y-6.0*y*y+4.0*y*y*y);
	values[1] = -y*y*(1.0-y)*(1.0-y)
	  *(2.0*x-6.0*x*x+4.0*x*x*x);
	break;
      case 3:
	values[0] = sin(12*pi*x) * cos(12*pi*y);
	values[1] = -cos(12*pi*x) * sin(12*pi*y);
	break;
      case 4: {
	real sin2 = sin(pi*x) * sin(pi*x);
	values[0] = 2*sin2*((1 - y)*(1 - y)*y - (1 - y)*y2);
	values[1] = -2*pi*cos(pi*x)*sin(pi*x)*(1 - y)*(1 - y)*y2;
	break;
      }
      case 5: {
	//	values[0] = 3*x3*y2;
	//	values[1] = -3*x2*y3;
	x = x*(1 - _perturb_x);
	y = y*(1 - _perturb_y);
	values[0] = ((1 - x)*x*(1 - y) - (1 - x)*x*y);
	values[1] = x*(1 - y)*y - (1 - x)*(1 - y)*y;
	break;
      }
      case 6: {
	real sin2 = sin(x)*sin(x);
	real pi_y2 = (pi - y) * (pi - y);
	// 2*sin(x)^2*(pi - y)^2*y - 2*sin(x)^2*(pi - y)*y^2
	// -2*cos(x)*sin(x)*(pi - y)^2*y^2

	values[0] = 2*sin2*(pi_y2*y - (pi - y)*y2);
	values[1] = -2*cos(x)*sin(x)*pi_y2*y2;
	break;
      }
      default:
	message("Unknown problem in Velocity, given %d",_problem);
	std::exit(1);
      }
    }
  };

  // -----------------------------------------------------------------------------
  // Defines cross derivatives [du_x/dy, du_y/dx]
  class TestVelocityDerivs : public D2R1Function
  {
  private:
    int _problem;
  public:
  TestVelocityDerivs(Mesh& mesh,int problem) : D2R1Function(mesh),_problem(problem)
    {}
    void eval(real* values, const real* x) const
    {
      switch (_problem) {
      case 1:
	values[0] = -3*DOLFIN_PI*sin(3*DOLFIN_PI*x[0]) * sin(3*DOLFIN_PI*x[1]);
	values[1] = 3*DOLFIN_PI*sin(3*DOLFIN_PI*x[0]) * sin(3*DOLFIN_PI*x[1]);
	break;
      case 2:
	values[0] = x[0]*x[0]*(1.0-x[0])*(1.0-x[0])
	  *(2.0-12.0*x[1]+12.0*x[1]*x[1]);
	values[1] = -x[1]*x[1]*(1.0-x[1])*(1.0-x[1])
	  *(2.0-12.0*x[0]+12.0*x[0]*x[0]);
	break;
      case 3:
	values[0] = -12*DOLFIN_PI*sin(12*DOLFIN_PI*x[0]) * sin(12*DOLFIN_PI*x[1]);
	values[1] = 12*DOLFIN_PI*sin(12*DOLFIN_PI*x[0]) * sin(12*DOLFIN_PI*x[1]);
	break;
      case 4:
	message("Test Velocity Derivs for Problem 4 unimplemented");
	std::exit(1);
	break;
      default:
	message("Unknown problem in VelocityDerivs, given %d",_problem);
	std::exit(1);      
      }
    }
  };

  // -----------------------------------------------------------------------------
  class TestVelocityCurl : public Function
  {
  private:
    int _problem;
  public:
  TestVelocityCurl(Mesh& mesh,int problem) : Function(mesh),_problem(problem)
    {}
    real eval(const real* x) const
    {
      TestVelocityDerivs dv(mesh(),_problem);
      real dv_values[2];
      dv.eval(dv_values,x);
      return dv_values[0]-dv_values[1];
    }
    virtual uint rank() const {return 0;}
    virtual uint dim() const {return 1;}
  };
  // ----------------------------------------------------------------------------
  class TestZ : public Function
  {
  private:
    int _problem;
    real _alpha;
    real _mu;
  public:
  TestZ(Mesh& mesh, int problem, real alpha=0, real mu=1.0) :
    Function(mesh), _problem(problem), _alpha(alpha), _mu(mu)
    {}
    uint rank() const {return 0;};
    real eval(const real* input) const
    {
      real value;
      real x = input[0];
      real x2 = x*x;
      //real x3 = x*x2;
      real y = input[1];
      real y2 = y*y;
      real y3 = y*y2;
      real y4 = y*y3;
      real pi = DOLFIN_PI;
      real pi2 = pi*pi;
      switch (_problem) {
      case 4:
	{
	  real sin2 = sin(pi*x)*sin(pi*x);
	  real cos2 = cos(pi*x)*cos(pi*x);
	  real y_1 = (1 - y);
	  real y_12 = y_1*y_1;
	  value = 2*_alpha*(-4*pi2*sin2*y2 + 4*pi2*cos2*y2 + 16*pi2*sin2*y_1*y - 
		       16*pi2*cos2*y_1*y - 4*pi2*sin2*y_12 + 4*pi2*cos2*y_12) + 
	    2*pi2*sin2*y_12*y2 - 2*pi2*cos2*y_12*y2 - 2*sin2*y2 + 
	    8*sin2*y_1*y - 2*sin2*y_12;
	  //	  value = 8*pi2*_alpha*((cos2 - sin2) + 4*(sin2 - cos2)*y_1*y) +
	  //	    2*pi2*(sin2 - cos2)*y_12*y2 - 2*sin2*(y2 + 4*y_1*y - y_12);
	}
      case 5: 
	{
	  //	  value = (72*_alpha*x - 6*x3)*y - 6*x*y3;
	  value = -2*y2 + 2*y - 2*x2 + 2*x + 8*_alpha;
	  break;
	}
      case 6:
	{
	  real sin2 = sin(x)*sin(x);
	  real cos2 = cos(x)*cos(x);
	  // (2*sin(x)^2 - 2*cos(x)^2)*y^4 + (4*pi*cos(x)^2 - 4*pi*sin(x)^2)*y^3 + 
	  // ((-48*a + 2*pi^2 - 12)*sin(x)^2 + (48*a - 2*pi^2)*cos(x)^2)*y^2 + 
	  // ((48*pi*a + 12*pi)*sin(x)^2 - 48*pi*a*cos(x)^2)*y + 
	  //(-8*pi^2*a - 2*pi^2)*sin(x)^2 + 8*pi^2*a*cos(x)^2
	  value =(2*sin2 - 2*cos2)*y4 + (4*pi*cos2 - 4*pi*sin2)*y3 + 	    
	    ((-48*_alpha + 2*pi2 - 12)*sin2 + (48*_alpha - 2*pi2)*cos2)*y2 +
	    ((48*pi*_alpha + 12*pi)*sin2 - 48*pi*_alpha*cos2)*y + 
	    (-8*pi2*_alpha - 2*pi2)*sin2 + 8*pi2*_alpha*cos2;
	  break;
	}
      default:
	message("Unknown problem in Z, given %d",_problem);
	std::exit(1);      
      }      
      return value;
    }
  };
  // ----------------------------------------------------------------------------
  class TestRHS : public D2R1Function
  {
  private:
    int _problem;
    real _alpha;
    real _mu;
  public:
  TestRHS(Mesh& mesh,int problem, real alpha=0, real mu=1.0) : 
    D2R1Function(mesh), _problem(problem), _alpha(alpha), _mu(mu)
    {}
    void eval(real* values, const real* input) const
    {
      real pi = DOLFIN_PI;
      real x = input[0];
      real x2 = x*x;
      real x3 = x2*x;
      real x4 = x3*x;
      real y = input[1];
      real y2 = y*y;
      real y3 = y2*y;
      real y4 = y3*y;

      switch (_problem) {
      case 1:
	values[0] = 18.0 * pi * pi * 
	  sin(3*pi*x) * cos(3*pi*y) + 
	  3*pi*cos(3*pi*x)*sin(3*pi*y);
	values[1] = -18.0 * pi * pi * 
	  cos(3*pi*x) * sin(3*pi*y) +
	  3*pi*sin(3*pi*x)*cos(3*pi*y);
	break;
      case 2:
	values[0] = (12.0 - 24.0*y)*x4
	  + (-24.0 + 48.0*y)*x3
	  + (-48.0*y + 72.0*y2 - 48.0*y3 + 12.0)*x2
	  + (-2.0 + 24.0*y - 72.0*y2 + 48.0*y3)*x
	  + 1.0 - 4.0*x + 12.0*x2 - 8.0*x3;
	values[1] = (8.0 - 48.0*y + 48.0*y2)*x3
	  + (-12.0 + 72.0*y - 72.0*y2)*x2
	  + (4.0 - 24.0*y + 48.0*y2 - 48.0*y3 + 24.0*y4)*x
	  - 12.0*y2 + 24.0*y3 - 12.0*y4;
	break;
      case 3: 
	values[0] = 288.0 * pi * pi * 
	  sin(12.0*pi*x) * cos(12.0*pi*y) + 
	  12.0*pi*cos(12.0*pi*x)*sin(12.0*pi*y);
	values[1] = -288.0 * pi * pi * 
	  cos(12.0*pi*x) * sin(12.0*pi*y) +
	  12.0*pi*sin(12.0*pi*x)*cos(12.0*pi*y);
	break;
      case 4: {
	real pi2 = pi*pi;
	real sin2 = sin(pi*x)*sin(pi*x);
	real cos2 = cos(pi*x)*cos(pi*x);
	real cossin = cos(pi*x)*sin(pi*x);
	real y_1 = 1 - y;
	real y_12 = y_1 * y_1;

	real lap_u0 = 4*pi2*sin2*y_1*y2 - 4*pi2*cos2*y_1*y2 - 
	  4*pi2*sin2*y_12*y + 4*pi2*cos2*y_12*y;
	real lap_u1 = -4*pi*cossin*y2 + 16*pi*cossin*y_1*y - 
	  4*pi*cossin*y_12;

	real pdx = -sin(x)*cos(2*y);
	real pdy = -2*cos(x)*sin(2*y);

	TestVelocity u_fun(mesh(),_problem);
	real u[2];
	u_fun.eval(u,input);
	TestZ z_fun(mesh(),_problem,_alpha,_mu);
	real z = z_fun.eval(input);

	values[0] =-_mu*lap_u0 + pdx - z*u[1];
	values[1] = -_mu*lap_u1 + pdy + z*u[0];
	break;
      }
      case 5: {
	TestVelocity u_fun(mesh(),_problem);
	real u[2];
	u_fun.eval(u,input);
	TestZ z_fun(mesh(),_problem,_alpha,_mu);
	real z = z_fun.eval(input);
	//	real lap_u0 = 18*x*y2;
	//	real lap_u1 = -18*x2*y;
	real lap_u0 = 4*y - 2;
	real lap_u1 = 2 - 4*x;
	real pdx = y;
	real pdy = x;
	values[0] = -_mu*lap_u0 + pdx - z*u[1];
	values[1] = -_mu*lap_u1 + pdy + z*u[0];
	break;
      }
      case 6: {
	real sin2 = sin(x)*sin(x);
	real cos2 = cos(x)*cos(x);
	real pi_y = pi - y;
	real pi_y2 = pi_y*pi_y;

	TestVelocity u_fun(mesh(),_problem);
	real u[2];
	u_fun.eval(u,input);
	TestZ z_fun(mesh(),_problem,_alpha,_mu);
	real z = z_fun.eval(input);

	// 4*sin(x)^2*(pi - y)*y^2 - 4*cos(x)^2*(pi - y)*y^2 - 
	//   4*sin(x)^2*(pi - y)^2*y + 4*cos(x)^2*(pi - y)^2*y 
	real lap_u0 = 4*((sin2 - cos2)*pi_y*y2 + 
			 (cos2 - sin2)*pi_y2*y);
	// -4*cos(x)*sin(x)*y^2 + 16*cos(x)*sin(x)*(pi - y)*y - 
	// 4*cos(x)*sin(x)*(pi - y)^2
	real lap_u1 = 4*cos(x)*sin(x)*(4*pi_y*y - pi_y2 - y2);
	// -sin(x)*cos(2*y)
	real pdx = -sin(x)*cos(2*y);
	// -2*cos(x)*sin(2*y)
	real pdy = -2*cos(x)*sin(2*y);
	values[0] = -_mu*lap_u0 + pdx - z*u[1];
	values[1] = -_mu*lap_u1 + pdy + z*u[0];
	break;	
      }
      default:
	message("Unknown problem in RHS, given %d",_problem);
	std::exit(1);      
      }
    }
  };

  // -----------------------------------------------------------------------------
  class TestPressureBC : public D1R1Function
  {
  public:
  TestPressureBC(Mesh& mesh) : D1R1Function(mesh) {}
    real eval (const real* x) const
    {    
      return 0.0;
    }
  };

  // -----------------------------------------------------------------------------
  // Boundary Subdomains
  // -----------------------------------------------------------------------------
  class TestBoundary : public SubDomain
  {
  public:
  TestBoundary() : SubDomain() {}
    virtual bool inside (const real* x, bool on_boundary) const
    {
      return on_boundary;
    }
  };

  // -----------------------------------------------------------------------------
  class TestPenPoint : public SubDomain
  {
  public:
    TestPenPoint() : SubDomain() {}
    virtual bool inside (const real* x, bool on_boundary) const
    {
      bool ret_val = fabs(x[0]) < DOLFIN_EPS && fabs(x[1]) < DOLFIN_EPS;
      return on_boundary && ret_val;  
    }
  };

}
#endif
