// Copyright (C) 2008 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.

#include "Grade2Solver.h"

using namespace dolfin;

// ----------------------------------------------------------------------------
Grade2Solver::Grade2Solver () :
  StokesSolver(), _z(0), f_alpha(0), f_mu(0), f_delta(0), f_epsilon(0), 
  f_norm(0), f_rho(0), f_w(0), f_h(0), _a_transport(0), _L_transport(0)
{
  // Set default parameters
  add("transport disc", "Galerkin");
  add("alpha", 1.0);
  add("mu", 1.0);
  add("max_iter",10);
  add("upwind param", 1e-2);
  add("regularize param", 1e-2);
}
// ----------------------------------------------------------------------------
Grade2Solver::Grade2Solver (const Grade2Solver& that)
{
  throw std::runtime_error("Copy constructer is not implemented.");
}
// ----------------------------------------------------------------------------
Grade2Solver::~Grade2Solver ()
{
  deletePtrs();
}
// ----------------------------------------------------------------------------
Grade2Solver& Grade2Solver::operator= (const Grade2Solver& that)
{
  throw std::runtime_error("Assignment is not implemented.");
  return *this;
}
// ----------------------------------------------------------------------------
void Grade2Solver::formFactory (Fluid& fluid, Function& f)
{
  // Remove old forms and functions
  StokesSolver::deletePtrs();
  deletePtrs();

  // Create necessary functions
  f_alpha = new Function(fluid.mesh(), get("alpha"));
  f_mu = new Function(fluid.mesh(), get("mu"));
  _z = new Function(fluid.mesh(),0.0);

  // Create the forms
  const std::string vel_disc = get("velocity disc");
  const std::string z_disc = get("transport disc");
  if (vel_disc == "Taylor-Hood") {
    _a_stokes = new P2THG2VelBilinearForm(*_z, *f_mu);
    _L_stokes = new P2THG2VelLinearForm(f);
    if (z_disc == "Galerkin") {
      _a_transport = new P2THG2GalZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha);
      _L_transport = new P2THG2GalZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f);
    } else if (z_disc == "symmetric") {
      _a_transport = new P2THG2SymZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha);
      _L_transport = new P2THG2SymZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f);
    } else if (z_disc == "streamline") {
      f_delta = new Function(fluid.mesh(), get("upwind param"));
      _a_transport = new P2THG2StrZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha, *f_delta);
      _L_transport = new P2THG2StrZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f, *f_delta);
    } else if (z_disc == "symmetric streamline") {
      f_delta = new Function(fluid.mesh(), get("upwind param"));
      _a_transport = new P2THG2SymStrZBilinearForm(fluid.velocity(),
						   *f_mu, *f_alpha, *f_delta);
      _L_transport = new P2THG2SymStrZLinearForm(fluid.velocity(),
						 *f_mu, *f_alpha, f, *f_delta);
    } else if (z_disc == "regularized") {
      f_epsilon = new Function(fluid.mesh(), get("regularize param"));
      f_norm = new FacetNormal(fluid.mesh());
      f_h = new MeshSize(fluid.mesh());
      _a_transport = new P2THG2RegZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha, 
						*f_epsilon, *f_norm,
						*f_h);
      _L_transport = new P2THG2RegZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f);
    } else {
      throw std::runtime_error("Unimplemented transport discretization");
    }
  } else if (vel_disc == "Scott-Vogelius") {
    f_rho = new Function(fluid.mesh(),get("SV penalty rho"));
    f_w = new Function;
    _a_stokes = new P4SVG2VelBilinearForm(*_z, *f_mu, *f_rho);
    _L_stokes = new P4SVG2VelLinearForm(f, *f_w);
    _a_div_proj = new P4VecDivBilinearForm;
    _L_div_proj = new P4VecDivLinearForm(*f_w);
    if (z_disc == "Galerkin") {
      _a_transport = new P4SVG2GalZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha);
      _L_transport = new P4SVG2GalZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f);
    } else if (z_disc == "symmetric") {
      _a_transport = new P4SVG2SymZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha);
      _L_transport = new P4SVG2SymZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f);
    } else if (z_disc == "streamline") {
      f_delta = new Function(fluid.mesh(), get("upwind param"));
      _a_transport = new P4SVG2StrZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha, *f_delta);
      _L_transport = new P4SVG2StrZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f, *f_delta);
    } else if (z_disc == "symmetric streamline") {
      f_delta = new Function(fluid.mesh(), get("upwind param"));
      _a_transport = new P4SVG2SymStrZBilinearForm(fluid.velocity(),
						   *f_mu, *f_alpha, *f_delta);
      _L_transport = new P4SVG2SymStrZLinearForm(fluid.velocity(),
						 *f_mu, *f_alpha, f, *f_delta);
    } else if (z_disc == "regularized") {
      f_epsilon = new Function(fluid.mesh(), get("regularize param"));
      f_norm = new FacetNormal(fluid.mesh());
      f_h = new MeshSize(fluid.mesh());
      _a_transport = new P4SVG2RegZBilinearForm(fluid.velocity(),
						*f_mu, *f_alpha, 
						*f_epsilon, *f_norm,
						*f_h);
      _L_transport = new P4SVG2RegZLinearForm(fluid.velocity(),
					      *f_mu, *f_alpha, f);
    } else {
      throw std::runtime_error("Unimplemented transport discretization");
    }
  } else {
    throw  std::runtime_error("Unimplemented Grade2 velocity discretization");
  }
}
// ----------------------------------------------------------------------------
void Grade2Solver::solve (Fluid& fluid, Function& f, bool verbose)
{
  const std::string vel_disc = get("velocity disc");
  const std::string z_disc = get("transport disc");

  // Files for writing stuff
  File vel_file("results/velocity.pvd");
  File vel_diff_file("results/velocity_diff.pvd");
  File press_file("results/pressure.pvd");
  File press_diff_file("results/press_diff.pvd");
  File z_file("results/transport.pvd");
  File z_diff_file("results/transport_diff.pvd");

  // Generate the forms
  formFactory(fluid,f);

  // Create boundary conditions on the velocity and pressure
  Array<DirichletBC*> bcs;
  createBCs(fluid, bcs);

  // Make some constant and temporary functions
  Function v_old;
  Function z_old;
  real converged = 1.0;
  int max_iter = get("max_iter");
  Form* vel_l2_diff;
  Form* z_l2_diff;
  if (vel_disc == "Taylor-Hood") {
    vel_l2_diff = new P2VecL2DiffFunctional(fluid.velocity(),v_old);
    z_l2_diff = new P2L2DiffFunctional(*_z,z_old);
  } else if (vel_disc == "Scott-Vogelius") {
    vel_l2_diff = new P4VecL2DiffFunctional(fluid.velocity(),v_old);
    z_l2_diff = new P4L2DiffFunctional(*_z,z_old);
  } else {
    throw std::runtime_error("Undefined velocity difference form");
  }

  // Iterate between Stokes and Transport solvers until converged
  real norm_u0 = 1.0;
  real norm_z0 = 1.0;
  for (int z_iters = 0; z_iters < max_iter && converged > DOLFIN_SQRT_EPS; 
       z_iters++) {
    if (vel_disc == "Taylor-Hood") {
      mixedMethodSolve(fluid,bcs,f);
    } else if (vel_disc == "Scott-Vogelius") {
      scottVogeliusSolve(fluid,bcs,f,*f_w);
    } else {
      throw std::runtime_error("Undefined velocity solver");
    }
    LinearPDE pde_transport(*_a_transport,*_L_transport, fluid.mesh());
    pde_transport.solve(*_z);
    
    if (z_iters > 0) {
      real vel_diff = sqrt(fabs(assemble(*vel_l2_diff,fluid.mesh())))/norm_u0;
      message("Grade2Solver iteration %d: ||u-u_old||/||u_0|| = %.2e",
	      z_iters,vel_diff);
      real z_diff = sqrt(fabs(assemble(*z_l2_diff,fluid.mesh())))/norm_z0;
      message("Grade2Solver iteration %d: ||z-z_old||/||z_0|| = %.2e",
	      z_iters,z_diff);
      converged = vel_diff + z_diff;
      v_old.vector() = fluid.velocity().vector();
      z_old.vector() = _z->vector();
    } else {
      if (vel_disc == "Scott-Vogelius") {
	P4VecL2Functional p4_vec_l2(fluid.velocity());
	norm_u0 = sqrt(fabs(assemble(p4_vec_l2,fluid.mesh())));
	P4L2Functional p4_l2(*_z);
	norm_z0 = sqrt(fabs(assemble(p4_l2,fluid.mesh())));
	v_old.init(fluid.mesh(),*_a_stokes,0);
	v_old.vector().zero();
	z_old.init(fluid.mesh(),*_a_transport,0);
	z_old.vector().zero();
      } else {
	P2VecL2Functional p2_vec_l2(fluid.velocity());
	norm_u0 = sqrt(fabs(assemble(p2_vec_l2,fluid.mesh())));
	P2L2Functional p2_l2(*_z);
	norm_z0 = sqrt(fabs(assemble(p2_l2,fluid.mesh())));
	v_old = fluid.velocity();
	z_old = *_z;
      }
    }
    if (verbose) {
      // Write functions to file
      vel_file << fluid.velocity();
      press_file << fluid.pressure();
      z_file << *_z;

      // Write projection of error to file
      if (vel_disc == "Scott-Vogelius") {
	P4VecDiffBilinearForm a_v_diff;
	P4VecDiffLinearForm L_v_diff(fluid.velocity(), *_v_ex);
	LinearPDE pde_v_diff(a_v_diff, L_v_diff, fluid.mesh());
	Function f_v_diff;
	pde_v_diff.solve(f_v_diff);
	vel_diff_file << f_v_diff;
	
	P4DiffBilinearForm a_p_diff;
	P4DiffLinearForm L_p_diff(fluid.pressure(), *_p_ex);
	LinearPDE pde_p_diff(a_p_diff, L_p_diff, fluid.mesh());
	Function f_p_diff;
	pde_p_diff.solve(f_p_diff);
	press_diff_file << f_p_diff;

	P4DiffBilinearForm a_z_diff;
	P4DiffLinearForm L_z_diff(*_z, *_z_ex);
	LinearPDE pde_z_diff(a_z_diff, L_z_diff, fluid.mesh());
	Function f_z_diff;
	pde_z_diff.solve(f_z_diff);
	z_diff_file << f_z_diff;
      } else if (vel_disc == "Taylor-Hood") {
	P2VecDiffBilinearForm a_v_diff;
	P2VecDiffLinearForm L_v_diff(fluid.velocity(), *_v_ex);
	LinearPDE pde_v_diff(a_v_diff, L_v_diff, fluid.mesh());
	Function f_v_diff;
	pde_v_diff.solve(f_v_diff);
	vel_diff_file << f_v_diff;

	P1DiffBilinearForm a_p_diff;
	P1DiffLinearForm L_p_diff(fluid.pressure(), *_p_ex);
	LinearPDE pde_p_diff(a_p_diff, L_p_diff, fluid.mesh());
	Function f_p_diff;
	pde_p_diff.solve(f_p_diff);
	press_diff_file << f_p_diff;

	P2DiffBilinearForm a_z_diff;
	P2DiffLinearForm L_z_diff(*_z, *_z_ex);
	LinearPDE pde_z_diff(a_z_diff, L_z_diff, fluid.mesh());
	Function f_z_diff;
	pde_z_diff.solve(f_z_diff);
	z_diff_file << f_z_diff;
      }
    }
  }

  // Clean up the boundary conditions.
  destroyBCs(bcs);
}
// ----------------------------------------------------------------------------
void Grade2Solver::solve (Fluid& fluid, Function& f,
			  Function& v_ex, Function& p_ex, Function& z_ex)
{
  _v_ex = &v_ex;
  _p_ex = &p_ex;
  _z_ex = &z_ex;
  solve(fluid, f, true);
}

// ----------------------------------------------------------------------------
void Grade2Solver::deletePtrs ()
{
  if (_z) {delete _z; _z=0;}
  if (_a_transport) {delete _a_transport; _a_transport = 0;}
  if (_L_transport) {delete _L_transport; _L_transport = 0; }
  if (f_alpha) {delete f_alpha; f_alpha = 0;}
  if (f_mu) {delete f_mu; f_mu = 0;}
  if (f_delta) {delete f_delta; f_delta = 0;}
  if (f_epsilon) {delete f_epsilon; f_epsilon = 0;}
  if (f_norm) {delete f_norm; f_norm = 0;}
  if (f_rho) {delete f_rho; f_rho = 0;}
  if (f_w) {delete f_w; f_w = 0;}
  if (f_h) {delete f_h; f_h = 0;}
}
// ----------------------------------------------------------------------------
// end of file
