// Copyright (C) 2008 Andy R Terrel
// Licensed under the GNU LGPL Version 3.0 or later.

#include "StokesSolver.h"

using namespace dolfin;

// ----------------------------------------------------------------------------
StokesSolver::StokesSolver () : 
  _a_stokes(0), _L_stokes(0), _a_div_proj(0), _L_div_proj(0),
  _pinpoint(0), _pressurepin(0)
{ 
  // Set default parameters
  add("velocity disc","Taylor-Hood");
  add("SV penalty r",1.0e5);
  add("SV penalty rho",1.0e5);
  add("SV max_iter",100);
}
// ----------------------------------------------------------------------------
StokesSolver::StokesSolver (const StokesSolver& that)
{
  throw std::runtime_error("Copy constructer is not implemented.");
}
// ----------------------------------------------------------------------------
StokesSolver::~StokesSolver ()
{
  deletePtrs();
}
// ----------------------------------------------------------------------------
StokesSolver& StokesSolver::operator= (const StokesSolver& that)
{
  throw std::runtime_error("Assignment is not implemented.");
  return *this;
}
// ----------------------------------------------------------------------------
void StokesSolver::solve (Fluid& fluid, Function& f)
{
  // Get rid of old data, and get the new discretization method
  deletePtrs();
  const std::string disc = get("velocity disc");
  
  // Make boundary conditions, if it is a mixed method put 
  // the bc's on the velocity subsystem.
  Array<DirichletBC*> bcs;
  createBCs(fluid, bcs);

  // Set forms and call the particular solution for each method.
  if (disc == "Taylor-Hood") {
    _a_stokes = new P2THStokesBilinearForm;
    _L_stokes = new P2THStokesLinearForm(f);
    mixedMethodSolve(fluid, bcs, f);
  } else if (disc == "Scott-Vogelius") {
    Function f_rho(fluid.mesh(),get("SV penalty rho"));
    Function f_w;
    _a_stokes = new P4SVStokesBilinearForm(f_rho);
    _L_stokes = new P4SVStokesLinearForm(f, f_w);
    _a_div_proj = new P4VecDivBilinearForm;
    _L_div_proj = new P4VecDivLinearForm(f_w);
    scottVogeliusSolve(fluid, bcs, f, f_w);
  } else {
    throw std::runtime_error("Unimplemented Stokes discretization");
  }

  // Clean up the boundary conditions.
  destroyBCs(bcs);
}
// ----------------------------------------------------------------------------
void StokesSolver::createBCs (Fluid& fluid, Array<DirichletBC*>& bcs)
{
  const std::string disc = get("velocity disc");
  if (disc == "Taylor-Hood") {
    _vel_subsystem = SubSystem(0);
    _pres_subsystem = SubSystem(1);
    for (uint i = 0; i < fluid.vel_bc_funcs().size(); i++) 
      bcs.push_back( new DirichletBC(*fluid.vel_bc_funcs()[i], 
				     *fluid.vel_subdomains(),
				     i, _vel_subsystem) );
    _pinpoint = new PinPoint();
    _pressurepin = new PressurePin(fluid.mesh());
    // Add a pressure pinpoint to the boundary conditions
    bcs.push_back( new DirichletBC(*_pressurepin, fluid.mesh(), *_pinpoint, 
				   _pres_subsystem, pointwise) );
  } else if (disc == "Scott-Vogelius") {
    for (uint i = 0; i < fluid.vel_bc_funcs().size(); i++) 
      bcs.push_back( new DirichletBC(*(fluid.vel_bc_funcs()[i]), 
				     *fluid.vel_subdomains(), i) );
  } else {
    throw std::runtime_error("Boundary Conditions undefined for Stokes eqn");
  }
}
// ----------------------------------------------------------------------------
void StokesSolver::destroyBCs (Array<DirichletBC*>& bcs)
{
  for (uint i = 0; i < bcs.size(); i++)
    delete bcs[i];
}
// ----------------------------------------------------------------------------
void StokesSolver::mixedMethodSolve (Fluid& fluid, Array<DirichletBC*>& bcs, 
				     Function& f)
{
  // Solve the PDE
  LinearPDE pde_stokes(*_a_stokes, *_L_stokes, fluid.mesh(), bcs);
  pde_stokes.solve(fluid.velocity(),fluid.pressure());
}
// ----------------------------------------------------------------------------
void StokesSolver::scottVogeliusSolve (Fluid& fluid, Array<DirichletBC*>& bcs, 
				       Function& f, Function& f_w)
{
  const real r = get("SV penalty r");
  const int max_iter = get("SV max_iter");

  // Initialize temporary function w
  f_w.init(fluid.mesh(), *_L_stokes, 2);
  f_w.vector().zero();

  // Create pde and update until ||div(u)|| < eps
  LinearPDE pde_stokes(*_a_stokes, *_L_stokes, fluid.mesh(), bcs);
  P4VecDivL2Functional div_func(fluid.velocity());
  real div_u = 1.0;
  for (int i = 0; i < max_iter && div_u > DOLFIN_SQRT_EPS; i++) {
    pde_stokes.solve(fluid.velocity());
    f_w.vector().axpy(r,fluid.velocity().vector());
    div_u = sqrt(fabs(assemble(div_func,fluid.mesh())));
    message("Scott Vogelius iteration %d, ||div(u)|| = %.2e", i, div_u);
  }

  // Project div(w) to get the pressure
  LinearPDE div_proj(*_a_div_proj, *_L_div_proj, fluid.mesh());
  div_proj.solve(fluid.pressure());
}
// ----------------------------------------------------------------------------
void StokesSolver::deletePtrs ()
{
  if (_a_stokes) {delete _a_stokes; _a_stokes = 0;}
  if (_L_stokes) {delete _L_stokes; _L_stokes = 0;}
  if (_a_div_proj) {delete _a_div_proj; _a_div_proj = 0;}
  if (_L_div_proj) {delete _L_div_proj; _L_div_proj = 0;}
  if (_pinpoint) {delete _pinpoint; _pinpoint = 0;}
  if (_pressurepin) {delete _pressurepin; _pressurepin = 0;}
}
// ----------------------------------------------------------------------------
// end of file
